#include <windows.h>
#include <iostream>
#include <algorithm>

// STLSOFT
#include "comstl/error/exceptions.hpp"
#include "comstl/collections/collection_sequence.hpp"
#include "comstl/util/value_policies.hpp"

// Standard interop stuff
#pragma warning(push)
#pragma warning(disable: 4278)
#import "mscorlib.tlb" no_namespace raw_interfaces_only
#pragma warning(pop)

// The collection server
#ifdef _DEBUG
#import "../InteropCollectionServer/bin/Debug/InteropCollectionServer.tlb" no_namespace
#else
#import "../InteropCollectionServer/bin/Release/InteropCollectionServer.tlb" no_namespace
#endif

// The suggested new enumerator policy
#include "get_enumerator_policy.h"

// A COM collection sequence with the new enumerator policy
typedef comstl::collection_sequence
        <
            IEnumerable,
            IEnumVARIANT,
            _variant_t,
            comstl::VARIANT_policy,
            const _variant_t&,
            comstl::input_cloning_policy<IEnumVARIANT>,
            8,
            get_enumerator_policy<IEnumerable>
        > EnumerableSequence;

// Helper function for extracting an int from a VARIANT
int IntFromVariant(const VARIANT& v)
{
    return v.iVal;
}

int main(void)
{
    try
    {
        // Init COM
        ::CoInitialize(NULL);

        // Create a collection server
        ICollectionServerPtr spCollectionServer(__uuidof(CollectionServer));
        spCollectionServer->Size = 5;

        // Create a collection sequence for the collection
        EnumerableSequence Collection(IEnumerablePtr(spCollectionServer), true);

        // Output collection
        std::cout << "Here comes the collection: ";
        std::transform(Collection.begin(), Collection.end(), std::ostream_iterator<int>(std::cout, " "), IntFromVariant);

        // Cleanup COM
        ::CoUninitialize();
        return 0;
    }

    catch (_com_error& e)
    {
        std::cout << "COM Error: " << e.ErrorMessage() << std::endl;
        return 1;
    }

    catch (comstl::com_exception& e)
    {
        std::cout << "COM Exception: " << e.what() << std::endl;
        return 2;
    }

    catch (...)
    {
        std::cout << "Unknown Exception!" << std::endl;
    }
}

